/* This is heavily based on the work outlined in the statement below.
 * As such it is released under the same license and conditions.
 * No warranty is provided... blah blah
 * Date of modification/creation 2009-2010
 * Author (in part): Brian Starkey
 * 
 * Most of the commented out stuff is work that was Christian's, which 
 * I commented out to add my own functions etc
 * My code is by no means neat or elegant, but it works
 *
 * Parts also based on http://eeegadgets.blogspot.com/2009/07/usb-mail-notifier-final-revision.html
 */
/* Name: main.c
 * Project: hid-custom-rq example
 * Author: Christian Starkjohann
 * Creation Date: 2008-04-07
 * Tabsize: 4
 * Copyright: (c) 2008 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt), GNU GPL v3 or proprietary (CommercialLicense.txt)
 * This Revision: $Id: main.c 692 2008-11-07 15:07:40Z cs $
 */

/*
This example should run on most AVRs with only little changes. No special
hardware resources except INT0 are used. You may have to change usbconfig.h for
different I/O pins for USB. Please note that USB D+ must be the INT0 pin, or
at least be connected to INT0 as well.
We assume that an LED is connected to port B bit 0. If you connect it to a
different port or bit, change the macros below:
*/
#define LED_PORT_DDR        DDRB
#define LED_PORT_OUTPUT     PORTB
#define R_BIT            1
#define G_BIT            3
#define B_BIT            4
#define G_PORT 1 << G_BIT
#define B_PORT 1 << B_BIT
#define R_PORT 1 << R_BIT
#define R_INDEX 0
#define G_INDEX 1
#define B_INDEX 2
#define S_INDEX 3

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>  /* for sei() */
#include <util/delay.h>     /* for _delay_ms() */

#include <avr/pgmspace.h>   /* required by usbdrv.h */
extern "C"
{
	#include "usbdrv.h"
}
#include "oddebug.h"        /* This is also an example for using debug macros */
#include "requests.h"       /* The custom request numbers we use */

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

PROGMEM char usbHidReportDescriptor[22] = {    /* USB report descriptor */
    0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0xc0                           // END_COLLECTION
};
/* The descriptor above is a dummy only, it silences the drivers. The report
 * it describes consists of one byte of undefined data.
 * We don't transfer our data through HID reports, we use custom requests
 * instead.
 */

/*static uint8_t count = 0;
static volatile uint8_t r = 0;
static volatile uint8_t g = 0;
static volatile uint8_t b = 0;
static uint8_t rb = 0;
static uint8_t gb = 0;
static uint8_t bb = 0;*/

unsigned char current[3] = {0,0,0};
unsigned char rgb[4] = {0,0,0,5}; //Red, Green, Blue, Speed
void tick (unsigned char length);
void move_to(unsigned char * newLevels);

//PWM routine, length is number of PWM cycles
void tick (unsigned char length) {
	
	unsigned char count, red = current[R_INDEX], green = current[G_INDEX], blue = current[B_INDEX];
	unsigned char op = 0b00011010;
	unsigned int i;
	
	for (i = 0; i < length; i++) {
		for (count = 0; count < 255; count++) {
				wdt_reset();		
				op = PORTB | 0b00011010;
				if (count < red) {
					op ^= R_PORT;
				}
				if (count < green) {
					op ^= G_PORT;
				}
				if (count < blue) {
					op ^= B_PORT;
				}
				PORTB = op;
				_delay_us(5);
				usbPoll();
		}
	}
}

//Move to the colour specified in the array 'newLevels'
void move_to(unsigned char * newLevels) {
	
	unsigned char red_n, green_n, blue_n, red_o, green_o, blue_o, speed;
	char red_d, green_d, blue_d;
	unsigned char done = 1;
	
	//New Levels
	red_n = newLevels[R_INDEX];
	green_n = newLevels[G_INDEX];
	blue_n = newLevels[B_INDEX];
	
	//Old Levels
	red_o = current[R_INDEX];
	green_o = current[G_INDEX];
	blue_o = current[B_INDEX];
	
	speed = rgb[S_INDEX];
	
	if (red_n >= red_o) {
		red_d = 1; }
	else {
		red_d = -1; }
	
	if (green_n >= green_o) {
		green_d = 1; }
	else {
		green_d = -1; }
		
	if (blue_n >= blue_o) {
		blue_d = 1; }
	else {
		blue_d = -1;}

	if (speed > 0) {	
		do {
		wdt_reset();	
		done = 1;
			if (red_o != red_n) {
				red_o += red_d;
				current[R_INDEX] = red_o;
				done = 0; }
			if (green_o != green_n) {
				green_o += green_d;
				current[G_INDEX] = green_o;
				done = 0; }
			if (blue_o != blue_n) {
				blue_o += blue_d; 
				current[B_INDEX] = blue_o;
				done = 0;}
			tick(speed);
		} while (!done);
	}
	else {
		current[R_INDEX] = red_n;
		current[G_INDEX] = green_n;
		current[B_INDEX] = blue_n;
		tick(10);
	}
}





/* ------------------------------------------------------------------------- */

extern "C" usbMsgLen_t usbFunctionSetup(uchar data[8])
{
	usbRequest_t    *rq = (usbRequest_t *)data;

    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_VENDOR)
	{
		switch(rq->bRequest)
		{
		case CUSTOM_RQ_SET_COLOUR:
			rgb[R_INDEX] = rq->wValue.bytes[0];
			rgb[G_INDEX] = rq->wValue.bytes[1];
			rgb[B_INDEX] = rq->wIndex.bytes[0];
			rgb[S_INDEX] = rq->wIndex.bytes[1];			
			//speed = rq->wIndex.bytes[1];			
			//eeprom_write_byte(0, OSCCAL);   // store the calibrated value in EEPROM
			//eeprom_write_block((void) *rgb, 1, 4); 		
			break;	
		case CUSTOM_RQ_SET_SPEED:
			rgb[S_INDEX] = rq->wValue.bytes[0];
			break;
		}    
		//move_to(rgb);
}
	else
	{
        /* class requests USBRQ_HID_GET_REPORT and USBRQ_HID_SET_REPORT are
         * not implemented since we never call them. The operating system
         * won't call them either because our descriptor defines no meaning.
         */
    }
    return 0;   /* default for not implemented requests: return no data back to host */
}

static void calibrateOscillator(void)
{
uchar       step = 128;
uchar       trialValue = 0, optimumValue;
int         x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);
 
    /* do a binary search: */
    do{
        OSCCAL = trialValue + step;
        x = usbMeasureFrameLength();    // proportional to current real frequency
        if(x < targetValue)             // frequency still too low
            trialValue += step;
        step >>= 1;
    }while(step > 0);
    /* We have a precision of +/- 1 for optimum OSCCAL here */
    /* now do a neighborhood search for optimum value */
    optimumValue = trialValue;
    optimumDev = x; // this is certainly far away from optimum
    for(OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++){
        x = usbMeasureFrameLength() - targetValue;
        if(x < 0)
            x = -x;
        if(x < optimumDev){
            optimumDev = x;
            optimumValue = OSCCAL;
        }
    }
    OSCCAL = optimumValue;
}
 

extern "C" void usbEventResetReady(void)
{
    cli();  // usbMeasureFrameLength() counts CPU cycles, so disable interrupts.
    calibrateOscillator();
    sei();
    eeprom_write_byte(0, OSCCAL);   // store the calibrated value in EEPROM
}


int main(void)
{
	uchar   i;

    wdt_enable(WDTO_1S);
	//wdt_disable();    
	/* Even if you don't use the watchdog, turn it off here. On newer devices,
     * the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
     */
    /* RESET status: all port bits are inputs without pull-up.
     * That's the way we need D+ and D-. Therefore we don't need any
     * additional hardware initialization.
     */

    usbInit();
    usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
    i = 0;
    while(--i){             /* fake USB disconnect for > 250 ms */
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();
    LED_PORT_DDR |= _BV(R_BIT);   /* make the LED bit an output */
    LED_PORT_DDR |= _BV(G_BIT);   /* make the LED bit an output */
    LED_PORT_DDR |= _BV(B_BIT);   /* make the LED bit an output */
	
    sei();

    for(;;){                /* main event loop */
		wdt_reset();
		usbPoll();
		move_to(rgb);
    }
    return 0;
}

/* ------------------------------------------------------------------------- */
